from typing import List, Union

from zeno.classes.base import CamelModel


class Points2D(CamelModel):
    x: List[float]
    y: List[float]
    color: List[int]
    opacity: List[int]
    ids: List[str]
    domain: List[Union[int, float, str]]
    # "nominal" or "continuous" or "boolean"
    dataType: str
